#include "asterisk.h"
#include "asterisk/translate.h"
#include "asterisk/module.h"
#include "asterisk/format.h"
#include "asterisk/utils.h"

/* version 1.0, compatiblity overview as of May 2014 */

enum amr_attr_keys {
	AMR_ATTR_KEY_SAMP_RATE,
	AMR_ATTR_KEY_OCTET_ALIGN,
	AMR_ATTR_KEY_MODE_SET,
	AMR_ATTR_KEY_MODE_CHANGE_PERIOD,
	AMR_ATTR_KEY_MODE_CHANGE_NEIGHBOR,
	AMR_ATTR_KEY_MAX_RED,
}; /* based on include/asterisk/silk.h */

#define AST_MODULE "codec_amr"
/* based on res/res_format_attr_silk.c */

struct amr_attr {
	unsigned int samplerate;
	unsigned int octet_align;
	unsigned int mode_set:9;
	unsigned int mode_change_period;
	unsigned int mode_change_neighbor;
	int max_red;
}; /* based on http://tools.ietf.org/html/rfc4867#section-8.2 */

int amr_sdp_parse(struct ast_format_attr *format_attr, const char *attrib)
{	
	struct amr_attr *attr = (struct amr_attr *) format_attr;
	char *attributes;
	char *tmp;
	unsigned int val;
	const unsigned int size = 9; /* same as bit-field definition of mode_set */
	int v[size];
	
	/* init each slot as 'not specified' */
	attributes = ast_strdupa(attrib);
	for (val = 0; val < size; val = val + 1) {
		v[val] = -1;
	}
	/* lower-case everything, so we are case-insensitive */
	/* no implemmentation is known which is affacted by this */
	for (tmp = attributes; *tmp; ++tmp) {
		*tmp = tolower(*tmp);
	} /* based on channels/chan_sip.c:process_a_sdp_image() */
	
	tmp = strstr(attributes, "octet-align=");
	if (NULL != tmp) {
		if (sscanf(tmp, "octet-align=%30u", &val) == 1) {
			attr->octet_align = val;
		}
	}
	
	tmp = strstr(attributes, "mode-set=");
	if (NULL != tmp) {
		if (sscanf(tmp, "mode-set=%30u,%30u,%30u,%30u,%30u,%30u,%30u,%30u,%30u",
				   &v[0], &v[1], &v[2], &v[3], &v[4], &v[5], &v[6], &v[7], &v[8]) > 0) {
			for (val = 0; val < size; val = val + 1) {
				if (0 <= v[val] && v[val] < size) {
					attr->mode_set = (attr->mode_set | (1 << v[val]));
				}
			}
		}
	}
	
	tmp = strstr(attributes, "mode-change-period=");
	if (NULL != tmp) {
		if (sscanf(tmp, "mode-change-period=%30u", &val) == 1) {
			attr->mode_change_period = val;
		}
	}
	
	tmp = strstr(attributes, "mode-change-neighbor=");
	if (NULL != tmp) {
		if (sscanf(tmp, "mode-change-neighbor=%30u", &val) == 1) {
			attr->mode_change_neighbor = val;
		}
	}
	
	tmp = strstr(attributes, "max-red=");
	if (NULL != tmp) {
		if (sscanf(tmp, "max-red=%30u", &val) == 1) {
			attr->max_red = val;
		}
	}
	
	return 0;
}

void amr_sdp_generate(const struct ast_format_attr *format_attr, unsigned int payload, struct ast_str **str)
{
	struct amr_attr *attr = (struct amr_attr *) format_attr;
	int appended = 0;
	int listed = 0;
	
	if (0 != attr->octet_align) {
		if (0 == appended) {
			ast_str_append(str, 0, "a=fmtp:%d ", payload);
		} else {
			ast_str_append(str, 0, "; ");
		}
		ast_str_append(str, 0, "octet-align=%d", attr->octet_align);
		appended = appended + 1;
	}
	if (0 != attr->mode_set)
	{
		if (0 == appended) {
			ast_str_append(str, 0, "a=fmtp:%d ", payload);
		} else {
			ast_str_append(str, 0, "; ");
		}
		ast_str_append(str, 0, "mode-set=");
		if (attr->mode_set & 0x01) {
			if (0 == listed) {
				ast_str_append(str, 0, "0");
			} else {
				ast_str_append(str, 0, ",0");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x02) {
			if (0 == listed) {
				ast_str_append(str, 0, "1");
			} else {
				ast_str_append(str, 0, ",1");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x04) {
			if (0 == listed) {
				ast_str_append(str, 0, "2");
			} else {
				ast_str_append(str, 0, ",2");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x08) {
			if (0 == listed) {
				ast_str_append(str, 0, "3");
			} else {
				ast_str_append(str, 0, ",3");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x10) {
			if (0 == listed) {
				ast_str_append(str, 0, "4");
			} else {
				ast_str_append(str, 0, ",4");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x20) {
			if (0 == listed) {
				ast_str_append(str, 0, "5");
			} else {
				ast_str_append(str, 0, ",5");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x40) {
			if (0 == listed) {
				ast_str_append(str, 0, "6");
			} else {
				ast_str_append(str, 0, ",6");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x80) {
			if (0 == listed) {
				ast_str_append(str, 0, "7");
			} else {
				ast_str_append(str, 0, ",7");
			}
			listed = listed + 1;
		}
		if (attr->mode_set & 0x100) {
			if (0 == listed) {
				ast_str_append(str, 0, "8");
			} else {
				ast_str_append(str, 0, ",8");
			}
			listed = listed + 1;
		}
		appended = appended + 1;
	}
	if (0 != attr->mode_change_period) {
		if (0 == appended) {
			ast_str_append(str, 0, "a=fmtp:%d ", payload);
		} else {
			ast_str_append(str, 0, "; ");
		}
		ast_str_append(str, 0, "mode-change-period=%d", attr->mode_change_period);
		appended = appended + 1;
	}
	if (0 != attr->mode_change_neighbor) {
		if (0 == appended) {
			ast_str_append(str, 0, "a=fmtp:%d ", payload);
		} else {
			ast_str_append(str, 0, "; ");
		}
		ast_str_append(str, 0, "mode-change-neighbor=%d", attr->mode_change_neighbor);
		appended = appended + 1;
	}
	if (0 <= attr->max_red) {
		if (0 == appended) {
			ast_str_append(str, 0, "a=fmtp:%d ", payload);
		} else {
			ast_str_append(str, 0, "; ");
		}
		ast_str_append(str, 0, "max-red=%d", attr->max_red);
		appended = appended + 1;
	}
	if (0 != appended) {
		ast_str_append(str, 0, "\r\n");
	}
	/* 
	 * AMR-WB, GSM gateway compatible setting:
	 * ast_str_append(str, 0, "a=fmtp:%d mode-set=0,1,2; mode-change-period=2; mode-change-neighbor=1\r\n", payload);
	 * less than 25 kb/s in SIP RTP which equals iLBC:30 in data traffic, and AMR-WB is a high definition codec!
	 */
}

/*
 * Octet Alignment
 *
 * AMR and AMR-WB support 'octet-align=1' which does not stuff the
 * header bits over octet borders and which is readable in
 * Wireshark. This parameter is negotiated via SDP. Belledonne and
 * CounterPath does not offer AMR, only AMR-WB.
 * 
 * Nokia Symbian/S60: configurable via user-interface; default: off
 * Nokia Series 40: configurable via OMA Client Provisioning only; default: off
 * Nokia Asha Software Platform: see Nokia Series 40
 * CounterPath Bria (iOS): no setting known; off
 * CounterPath Bria (Android): see iOS
 * CounterPath Bria (BlackBerry 10): AMR/AMR-WB not available, as of version 1.3.1
 * Belledonne Linphone 1.0.2 (Windows Phone 8): no setting known; on
 *    ignores octet-align=0 in fmtp = distorted audio
 * CSipSimple 1.02r2330 (Android): no setting known; off
 *    AMR-WB part ignores octet-align=1 in fmtp = distorted audio; AMR works
 *    Bug: CSipSimple calls a Nokia Symbian/S60, no audio in Nokia; AMR works
 */
static enum ast_format_cmp_res amr_cmp(const struct ast_format_attr *fattr1, const struct ast_format_attr *fattr2)
{
	struct amr_attr *attr1 = (struct amr_attr *) fattr1;
	struct amr_attr *attr2 = (struct amr_attr *) fattr2;
	enum ast_format_cmp_res res = AST_FORMAT_CMP_EQUAL;

	if (attr1->samplerate != attr2->samplerate) {
		return AST_FORMAT_CMP_NOT_EQUAL;
	}
	if (attr1->octet_align != attr2->octet_align) {
		res = AST_FORMAT_CMP_SUBSET;
	}
	if (0 != attr1->mode_set && 0 != attr2->mode_set) {
		if (attr1->mode_set != attr2->mode_set) {
			if (attr1->mode_set & attr2->mode_set) {
				res = AST_FORMAT_CMP_SUBSET;
			} else {
				/* not expected because everyone supports 0,1,2 */
				ast_log(LOG_WARNING, "mode-set does not match\n");
				return AST_FORMAT_CMP_NOT_EQUAL;
			}
		}
	} else if (0 != attr1->mode_set || 0 != attr2->mode_set) {
		/* one is not specified = 0 = all; is the other one specified for all? */
		if (16000 == attr1->samplerate) /* AMR-WB: 0..8 = 9 bits */ {
			if (0 == attr1->mode_set && 0x1ff != attr2->mode_set) {
				res = AST_FORMAT_CMP_SUBSET;
			}
			if (0 == attr2->mode_set && 0x1ff != attr1->mode_set) {
				res = AST_FORMAT_CMP_SUBSET;
			}
		} else {
			if (0 == attr1->mode_set && 0xff != attr2->mode_set) {
				res = AST_FORMAT_CMP_SUBSET;
			}
			if (0 == attr2->mode_set && 0xff != attr1->mode_set) {
				res = AST_FORMAT_CMP_SUBSET;
			}
		}
	}
	if (attr1->mode_change_period != attr2->mode_change_period) {
		res = AST_FORMAT_CMP_SUBSET;
	}
	if (attr1->mode_change_neighbor != attr2->mode_change_neighbor) {
		res = AST_FORMAT_CMP_SUBSET;
	}
	if (attr1->max_red != attr2->max_red) {
		res = AST_FORMAT_CMP_SUBSET;
	}
	/* res is ignored, because _SUBSET is not supported as of Asterisk 11 */
	return AST_FORMAT_CMP_EQUAL;
}

static int amr_get_val(const struct ast_format_attr *format_attr, int key, void *result)
{
	struct amr_attr *attr = (struct amr_attr *) format_attr;
	int *val = result;

	switch (key) {
	case AMR_ATTR_KEY_SAMP_RATE:
		*val = attr->samplerate;
		break;
	case AMR_ATTR_KEY_OCTET_ALIGN:
		*val = attr->octet_align;
		break;
	case AMR_ATTR_KEY_MODE_SET:
		*val = attr->mode_set;
		break;
	case AMR_ATTR_KEY_MODE_CHANGE_PERIOD:
		*val = attr->mode_change_period;
		break;
	case AMR_ATTR_KEY_MODE_CHANGE_NEIGHBOR:
		*val = attr->mode_change_neighbor;
		break;
	case AMR_ATTR_KEY_MAX_RED:
		*val = attr->max_red;
		break;
	default:
		ast_log(LOG_WARNING, "unknown attribute type %d\n", key);
		return -1;
	}
	return 0;
}

static int amr_isset(const struct ast_format_attr *format_attr, va_list ap)
{
	enum amr_attr_keys key;
	struct amr_attr *attr = (struct amr_attr *) format_attr;

	for (key = va_arg(ap, int);
		key != AST_FORMAT_ATTR_END;
		key = va_arg(ap, int))
	{
		switch (key) {
		case AMR_ATTR_KEY_SAMP_RATE:
			if (attr->samplerate != (va_arg(ap, unsigned int))) {
				return -1;
			}
			break;
		case AMR_ATTR_KEY_OCTET_ALIGN:
			if (attr->octet_align != (va_arg(ap, unsigned int))) {
				return -1;
			}
			break;
		case AMR_ATTR_KEY_MODE_SET:
			if (attr->mode_set != (va_arg(ap, unsigned int))) {
				return -1;
			}
			break;
		case AMR_ATTR_KEY_MODE_CHANGE_PERIOD:
			if (attr->mode_change_period != (va_arg(ap, unsigned int))) {
				return -1;
			}
			break;
		case AMR_ATTR_KEY_MODE_CHANGE_NEIGHBOR:
			if (attr->mode_change_neighbor != (va_arg(ap, unsigned int))) {
				return -1;
			}
			break;
		case AMR_ATTR_KEY_MAX_RED:
			if (attr->max_red != (va_arg(ap, int))) {
				return -1;
			}
			break;
		default:
			ast_log(LOG_WARNING, "unknown attribute type %d\n", key);
			return -1;
		}
	}
	return 0;
}

static int amr_getjoint(const struct ast_format_attr *fattr1, const struct ast_format_attr *fattr2, struct ast_format_attr *result)
{
	struct amr_attr *attr1 = (struct amr_attr *) fattr1;
	struct amr_attr *attr2 = (struct amr_attr *) fattr2;
	struct amr_attr *attr_res = (struct amr_attr *) result;	

	if (attr1->samplerate != attr2->samplerate) {
		return -1;
	}
	
	attr_res->samplerate = attr1->samplerate; /* both same, use one */
	attr_res->octet_align = MAX(attr1->octet_align, attr2->octet_align);
	
	if (0 == attr1->mode_set && 0 == attr2->mode_set) {
		attr_res->mode_set = 0; /* both allowed all = 0 */
	} else if (0 != attr1->mode_set && 0 == attr2->mode_set) {
		attr_res->mode_set = attr1->mode_set; /* attr2 allowed all */
	} else if (0 == attr1->mode_set && 0 != attr2->mode_set) {
		attr_res->mode_set = attr2->mode_set; /* attr1 allowed all */
	} else { /* both parties restrict, let us check if they match */
		attr_res->mode_set = (attr1->mode_set & attr2->mode_set);
		if (0 == attr_res->mode_set) {
			/* not expected because everyone supports 0,1,2 */
			ast_log(LOG_WARNING, "mode-set does not match\n");
			return -1;
		}
	}
	attr_res->mode_change_period = MAX(attr1->mode_change_period, attr2->mode_change_period);
	attr_res->mode_change_neighbor = MAX(attr1->mode_change_neighbor, attr2->mode_change_neighbor);
	attr_res->max_red = MAX(attr1->max_red, attr2->max_red);

	return 0;
}

static void amr_set(struct ast_format_attr *format_attr, va_list ap)
{
	enum amr_attr_keys key;
	struct amr_attr *attr = (struct amr_attr *) format_attr;

	for (key = va_arg(ap, int);
		key != AST_FORMAT_ATTR_END;
		key = va_arg(ap, int))
	{
		switch (key) {
		case AMR_ATTR_KEY_SAMP_RATE:
			attr->samplerate = (va_arg(ap, unsigned int));
			break;
		case AMR_ATTR_KEY_OCTET_ALIGN:
			attr->octet_align = (va_arg(ap, unsigned int));
			break;
		case AMR_ATTR_KEY_MODE_SET:
			attr->mode_set = (va_arg(ap, unsigned int));
			break;
		case AMR_ATTR_KEY_MODE_CHANGE_PERIOD:
			attr->mode_change_period = (va_arg(ap, unsigned int));
			break;
		case AMR_ATTR_KEY_MODE_CHANGE_NEIGHBOR:
			attr->mode_change_neighbor = (va_arg(ap, unsigned int));
			break;
		case AMR_ATTR_KEY_MAX_RED:
			attr->max_red = (va_arg(ap, int));
			break;
		default:
			ast_log(LOG_WARNING, "unknown attribute type %d\n", key);
		}
	}
}

int amr_get_samples(const struct ast_frame *f)
{
	const struct ast_format *format = &f->subclass.format;
	unsigned int value = ast_format_dynamic_get_rate(format);
	
	if (0 != value) {
		return (value/50);
	}
	return 320; /* fallback, if not specified; AMR = 160 */
}

int amr_get_rate(const struct ast_format *format)
{
	int result = 0;
	const int key = AMR_ATTR_KEY_SAMP_RATE;
	unsigned int value = 0;
	
	result = ast_format_get_value(format, key, &value);
	if (0 == result && 0 != value) {
		return value;
	}
	return 16000; /* fallback, if not specified; AMR = 8000 */
}

int amr_allowSmoother(void)
{
	return 0; /* do not allow a smoother */
}

static struct ast_format_attr_interface amr_interface = {
	.id = 0, /* allocate dynamically; in Asterisk 11, yields to 100030 */
	.format_attr_cmp = amr_cmp,
	.format_attr_get_joint = amr_getjoint,
	.format_attr_set = amr_set,
	.format_attr_isset = amr_isset,
	.format_attr_get_val = amr_get_val,
	.format_attr_sdp_parse = amr_sdp_parse,
	.format_attr_sdp_generate = amr_sdp_generate,
	
	.format_samples	= amr_get_samples,
	.format_rate = amr_get_rate,
	.allowSmoother = amr_allowSmoother,
};

static int unload_module(void)
{
	ast_log(LOG_WARNING, "Unload was not tested yet, use with care!\n"); /* TODO */
	return ast_format_attr_unreg_interface(&amr_interface);
}

static int register_audio_attributes(enum ast_format_id id,
	const char *name,
	const char *subtype,
	const unsigned int samplespersecond)
{
	const char *desc = subtype;
	const int map = -1; /* payload ID in SDP, here: allocate dynamically */
	const int rtp_code = 0; /* ignored because parameter 'format' is set */
	struct ast_format_list entry;
	memset(&entry, 0, sizeof(entry));
	
	/* based on main/format.c:custom_silk_format */
	ast_copy_string(entry.name, name, sizeof(entry.name));
	entry.samplespersecond = samplespersecond;
	ast_copy_string(entry.desc, desc, sizeof(entry.desc));
	entry.fr_len = 20; /* ignored because we do not use a smoother */
	entry.min_ms = 20;
	entry.max_ms = 20;
	entry.inc_ms = 20;
	entry.def_ms = 20;
	entry.custom_entry = 0; /* static format, dynamically added */
	/* entry.custom_entry = 1; custom format (see codecs.conf) */
	/* TODO 'custom formats' are not added after restart of Asterisk */
	
	ast_format_set(&entry.format, id, 1,
		AMR_ATTR_KEY_SAMP_RATE, entry.samplespersecond, /* AMR or AMR-WB */
		AMR_ATTR_KEY_OCTET_ALIGN, 0,          /* default = off           */
		AMR_ATTR_KEY_MODE_SET, 0,             /* default = all           */
		AMR_ATTR_KEY_MODE_CHANGE_PERIOD, 0,   /* default = not specified */
		AMR_ATTR_KEY_MODE_CHANGE_NEIGHBOR, 0, /* default = change to any */
		AMR_ATTR_KEY_MAX_RED, -1,  /* redunancy, default = no limit      */
		AST_FORMAT_ATTR_END);
	/* based on main/rtp_engine.c:ast_rtp_engine_load_format */
	
	ast_rtp_format_dynamic_add(&entry.format, rtp_code,
		"audio", subtype,
		entry.samplespersecond, map);
	
	return ast_format_dynamic_add(&entry);
}

static int load_module(void)
{
	if (ast_format_dynamic_register(&amr_interface, AST_FORMAT_TYPE_AUDIO)) {
		return AST_MODULE_LOAD_DECLINE;
	}
	
	register_audio_attributes(amr_interface.id, "amrwb", "AMR-WB", 16000);
	register_audio_attributes(amr_interface.id, "amr", "AMR", 8000);
	
	/* TODO unload_module(); return AST_MODULE_LOAD_DECLINE; */
	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, 
	"AMR Audio Codec Module", 
	.load = load_module, 
	.unload = unload_module, 
	/*
	 * -1 adds custom formats AFTER us (currently SILK and CELT) because 
	 * main/format_pref.c:ast_codec_choose uses a list (array) which 
	 * is NOT synchronized with main/format.c:format_list (ao2_container; 
	 * used as linked list). Or stated differently: 
	 * main/format.c:load_format_config removes all custom formats which
	 * would move us down, and breaks format_pref; therefore -1
	 */
	.load_pri = AST_MODPRI_CHANNEL_DEPEND-1, 
);
